/*
	Algorytm radix-2 FFT przystosowany dla roznej dlugosci transformaty.
	
	Obliczenia staloprzecinkowe (fractional).
	
	Do rewersji binarnej (mod_value):
	0x00400000 <-> 0x00000200 - 1024 punkty
	0x10000000 <-> 0x00000008 - 16 punktow 
	0x04000000 <-> 0x00000020 - 64 punkty

*/


#include "def21161.h"
#include "asm_sprt.h"

   
#define N 64 				//dlugosc FFT
#define N_div_2 32 			//(dlugosc FFT)/2
#define log2N 6             //log2(dlugosc FFT)
#define nover2 32			//(dlugosc FFT)/2
#define nover4 16			//(dlugosc FFT)/4
#define Ntimes2 128			//(dlugosc FFT)*2
#define mod_value H#04000000//(dlugosc FFT)/2 przedstawiona w rewersji binarnej

// Deklaracje zmiennych globalnych
.GLOBAL inplacereal; 
.GLOBAL inplaceimag;
.GLOBAL inputreal;
.GLOBAL inputimag;
.GLOBAL groups;
.GLOBAL bflys_per_group;
.GLOBAL node_space;
.GLOBAL blk_exponent;

.GLOBAL twid_real;
.GLOBAL twid_imag;

//Deklaracje procedur wystepujacych w programie
.GLOBAL _main;
.GLOBAL _scramble;
.GLOBAL _fft_strt;

/*
Niezbedne do uruchomienia programu na plytce testowej

.EXTERN		RX_left_flag;

.EXTERN		Left_Channel_In0;
.EXTERN		Right_Channel_In0;
.EXTERN		Left_Channel_In1;
.EXTERN		Right_Channel_In1;

.EXTERN		Left_Channel_Out0;
.EXTERN		Right_Channel_Out0;

.EXTERN		Left_Channel_Out1;
.EXTERN		Right_Channel_Out1;
.EXTERN		Left_Channel_Out2;
.EXTERN		Right_Channel_Out2;

.EXTERN 	Left_Channel_AD1852;
.EXTERN 	Right_Channel_AD1852;
*/
   
.SECTION/DM seg_dmda;			//Segment pamieci danych

.VAR inplacereal [N];
.VAR inplaceimag [N];
.VAR padding[4] = 0,0,0,0;
.VAR inputreal [N];
.VAR inputimag [N];
.VAR groups = N_div_2;
.VAR bflys_per_group = 2;
.VAR node_space = 2;
.VAR blk_exponent = 0;

.SECTION/PM seg_pmda;			//Segment pamieci programu (dane)

.VAR twid_real [N_div_2] = "twid_real.dat";		//deklaracja tablic wspolczynnikow zespolonych
.VAR twid_imag [N_div_2] = "twid_imag.dat";		//inicjalizowanych z plikow .dat


.SECTION/PM seg_pmco;			//Segment pamieci programu (kod)

_main:							//Glowna procedura programu	

	CALL _scramble;
	CALL _fft_strt;
  	RTS;						//Powrot z procedury main

._main.END:

_scramble:						//Procedura zapisujaca probki wejsciowe w odwroconym porzadku (rewersja binarna)
	I1=inputreal; 				//I4 <- adres bufora z probkami wejsciowymi w normalnej kolejnosci
	I0=inplacereal; 			//I0 <- adres bufora z probkami zapisanymi w odwroconej kolejnosci
	BITREV(I0,0);				//Zapisanie w I0 adresu w rewersji binarnej
	M1=1;						//M1 <- modyfikator adresu
	M0=mod_value; 				//M0 <- modyfkator zapisany w rewersji binarnej
	L1=0;						//L1 i L0 ustawione na 0 -> bufory kolowe wylaczone
	L0=0;
	LCNTR = N;					//tyle iteracji petli, ile mamy probek
	BIT SET MODE1 BR0;			//Wlaczenie dla I0 w DAG1 trybu adresowania z rewersja binarna
	DO brev UNTIL LCE;
	R1=DM(I1,M1); 				//Odczyt probek wejsciowych po kolei
brev: 
	DM(I0,M0)=R1; 				//Zapis danych w odwroconej kolejnosci
	BIT CLR MODE1 BR0; 			//Wylaczenie dla I0 w DAG1 trybu adresowania z rewersja binarna
	RTS; 						//Powrot z procedury
	
._scramble.END:



_fft_strt:
	M0=0;
	M1=1;
	L1=0;
	L2=0;
	L3=0;
	B8=twid_real;				//Wpisanie do rejesrow bazowych B8 i B9 adresow tablic
	B9=twid_imag;				//ze wspolczynnikami zespolonymi
	L8=N_div_2; 				//Inicjalizacja buforow kolowych o dlugosci N/2
	L9=N_div_2;
	L10=0;
	BIT SET MODE1 CBUFEN;		//Umozliwienie korzystania z buforow kolowych
	R3=-2;
	//{ ETAP 1 }
	I0=inplacereal;				//Umieszczenie w rejestrach indeksowych I0-I3 adresow	
	I1=inplacereal + 1;			//bazowych buforow dla czesci rzeczywistej i urojonej 
	I2=inplaceimag;				//transformaty FFT. 
	I3=inplaceimag + 1;
	I8=twid_real;
	M8=0;
	M2=2;
	LCNTR=nover2;				//inicjalizacja N/2 obiegow petli
	R0=DM(I0,M0);
	R8=DM(I1,M0);
	R9=DM(I3,M0), R2=PM(I8,M8);	
	DO group_lp UNTIL LCE;
	MRF=R8*R2(SSFR);			//mnozenie dwoch liczb typu fractional z zaokragleniem w MAC
	R8=MR1F;
	R7=R0+R8, R1=DM(I2,M0);
	R7=ASHIFT R7 BY -1;			//skalowanie (dzielenie przez 2) celem unikniecia przepelnienia
 	DM(I0,M2)=R7;
	R7=R0-R8;
	R7=ASHIFT R7 BY -1;
	DM(I1,M2)=R7;
	MRF=R1*R2(SSFR);
	R1=MR1F;
	R7=R1+R9;
	R7=ASHIFT R7 BY -1;
	DM(I2,M2)=R7;
	R7=R1-R9, R0=DM(I0,M0);
 	R7=ASHIFT R7 BY -1;
	DM(I3,M2)=R7;
	R8=DM(I1,M0);
group_lp: 
	R9=DM(I3,M0);
	LCNTR=log2N - 2;				//inicjalizacja log2N-2 obiegow petli
	//{ETAPY OD 2 DO N-1}
	DO stage_loop UNTIL LCE; 		
	I0=inplacereal; 				//Umieszczenie w rejestrach I0 i I2 adresow
	I2=inplaceimag; 				//bazowych buforow czesci rzeczywistej i urojonej FFT
	R2=DM(groups);					//R2 <- liczba grup (liczba grup = N / (ETAP*2) )
	R2=ASHIFT R2 BY -1; 			//liczba grup = liczba grup /2
	DM(groups)=R2; 
	LCNTR=R2; 						//inicjalizacja groups obiegow petli
	M8=R2; 							//M8 <- modyfikator adresow dla wspolczynnikow zespolonych
	M2=DM(node_space); 				//M2 <- modyfikator "szerokosci skrzydelek" motylka
	I1=I0;
	MODIFY(I1,M2); 					//I1 <- I0 + szerokosc skrzydelek motylka
	I3=I2;
	MODIFY(I3,M2); 					//I3 <- I2 + szerokosc skrzydelek motylka
	DO group_loop UNTIL LCE;
	I8=twid_real; 					//I8 <- cos(2kpi/N)
	I9=twid_imag; 					//I9 <- -sin(2kpi/N)
	LCNTR=DM(bflys_per_group); 		//inicjalizacja liczba_motylkow obiegow petli
	R12=PM(I8,M8),R4=DM(I1,M0); 	//R12=cos(2kpi/N),R4=x1
	R13=PM(I9,M8),R5=DM(I3,M0); 	//R13=-sin(2kpi/N),R5=y1
	DO bfly_loop UNTIL LCE;
	MRF=R4*R13(SSFR),R0=DM(I0,M0); 		//MRF=x1*(-sin(2kpi/N)),R0=x0
	MRF=MRF+R5*R12(SSFR),R1=DM(I2,M0); 	//MRF=(y1*(cos(2kpi/N))+x1*(-sin(2kpi/N))),R1=y0
	R9=MR1F;							//R9=y1*(cos(2kpi/N))+x1*(-sin(2kpi/N))
	MRF=R4*R12(SSFR); 					//MRF=x1*(cos(2kpi/N))				
	MRF=MRF-R5*R13(SSFR); 				//MRF=x1*(cos(2kpi/N))-y1*(-sin(2kpi/N))
	R8=MR1F;							//R8=x1*(cos(2kpi/N))-y1*(-sin(2kpi/N))
	R7=R1-R9; 							//R7=y0-[y1*(cos(2kpi/N))+x1*(-sin(2kpi/N))]
	R7=ASHIFT R7 BY -1;					//skalowanie (dzielenie przez 2)
	DM(I3,M1)=R7; 						//zapis wyniku do bufora
	R7=R0-R8,R5=DM(I3,M0),R13=PM(I9,M8);//R7=x0-[x1*(cos(2kpi/N))-y1*(-sin(2kpi/N))], R5=nastepny(y1),R13=nastepny(-sin(2kpi/N))
	R7=ASHIFT R7 BY -1;
	DM(I1,M1)=R7; 						
	R7=R0+R8,R4=DM(I1,M0),R12=PM(I8,M8);//R7=x0+[x1*(cos(2kpi/N))-y1*(-sin(2kpi/N))], R4=nastepny(x1),R12=nastepny(cos(2kpi/N))
	R7=ASHIFT R7 BY -1;
	DM(I0,M1)=R7; 						
	R7=R1+R9; 							//{R7=y0+[y1*(cos(2kpi/N))+x1*(-sin(2kpi/N))]}
	R7=ASHIFT R7 BY -1;
bfly_loop: 
	DM(I2,M1)=R7; 						
	MODIFY(I0,M2); 						//modyfikacja I0,I1,I2,I3 o szerokosc motylka
	MODIFY(I1,M2); 						
	MODIFY(I2,M2); 						
group_loop: 
	MODIFY(I3,M2); 						
	R2=DM(node_space);
	R2=ASHIFT R2 BY 1;
	DM(node_space)=R2; 					//node_space=node_space*2 (zwiekszenie szerokosci motylka)
	R2=DM(bflys_per_group);
	R2=ASHIFT R2 BY 1;
stage_loop: 
	DM(bflys_per_group)=R2; 			//bflys_per_group=bflys_per_group*2 (zwiekszenie ilosci motylkow w grupie)
	//{ OSTATNI ETAP }
	I0=inplacereal;
	I1=inplacereal+nover2;
	I2=inplaceimag;
	I3=inplaceimag+nover2;
	LCNTR=N_div_2;
	M2=DM(node_space);
	M8=1;
	I8=twid_real;
	I9=twid_imag;
	R12=PM(I8,M8),R4=DM(I1,M0); 
	R13=PM(I9,M8),R5=DM(I3,M0); 
	DO bfly_lp UNTIL LCE;
	MRF=R4*R13(SSFR),R0=DM(I0,M0);
	MRF=MRF+R5*R12(SSFR),R1=DM(I2,M0);
	R9=MR1F;
	MRF=R4*R12(SSFR);
	MRF=MRF-R5*R13(SSFR);
	R8=MR1F;
	R7=R1-R9;
	R7=ASHIFT R7 BY -1;
	DM(I3,M1)=R7;
	R7=R0-R8,R5=DM(I3,M0),R13=PM(I9,M8);
	R7=ASHIFT R7 BY -1;
	DM(I1,M1)=R7;
	R7=R0+R8,R4=DM(I1,M0),R12=PM(I8,M8);
	R7=ASHIFT R7 BY -1;
	DM(I0,M1)=R7;
	R7=R1+R9;
	R7=ASHIFT R7 BY -1;
bfly_lp: 
	DM(I2,M1)=R7; 
	RTS;

._fft_strt.END:


